#!/usr/bin/perl

# Copyright Platform Computing Inc., an IBM company, 2012

# This is used both as a standalone utility as well as included
# by othe perl scripts who set $util=1 first to keep anything
# from running here.
if ($util != 1) {
  mkreport_main(\@ARGV);
}

#
#  avg / stddev / relative stderr
#
sub get_avg {
  my @a = @_;
  my $sum;
  my $d;
  $sum = 0;
  for $d (@a) {
    $sum += $d;
  }
  return $sum / ($#a + 1);
}
sub get_stddev {
  my @a = @_;
  my $sum;
  my $avg;
  my $d;

  $avg = get_avg(@a);
  $sum = 0;
  for $d (@a) {
    $sum += ($d - $avg) * ($d - $avg);
  }
  return sqrt($sum / ($#a + 1));
}
sub get_lowest {
  my @a = @_;
  my $tmp;
  my $d;
  $tmp = $a[0];
  for $d (@a) {
    if ($d < $tmp) { $tmp = $d; }
  }
  return $tmp;
}
sub get_highest {
  my @a = @_;
  my $tmp;
  my $d;
  $tmp = $a[0];
  for $d (@a) {
    if ($d > $tmp) { $tmp = $d; }
  }
  return $tmp;
}

#
#  equivalent to the C function by the same name
#  creates a .dat and .datinfo file
#
sub print_plot2d {
  my $nlines = shift @_;
  my @useline = @{shift @_};
  my $nitems = shift @_;
  my @data = @{shift @_};
  my $lower_is_better = shift @_;
  my $percentage_away_to_flag = shift @_;
  my $xlabel = shift @_;
  my @xlabels = @{shift @_};
  my $ylabel = shift @_;
  my @ylabels = @{shift @_};
  my $graph_title = shift @_;
  my $filename_string = shift @_;

  my $i, $j;
  my $filename;
  my $fp;
  my $tmp;

  $filename = sprintf("out.%s.datinfo", $filename_string);
  open $fp, "> $filename";
  print $fp "2d\n";
  print $fp "$filename_string\n";
  print $fp "$graph_title\n";
  print $fp "misc info\n";
  print $fp "$lower_is_better\n";
  printf $fp "%.4f\n", $percentage_away_to_flag;
  print $fp "y axis info\n";
  print $fp "$ylabel\n";
  print $fp "$nlines\n";
  for ($i=0; $i<$nlines; ++$i) {
    print $fp "$ylabels[$i]\n";
    print $fp "$useline[$i]\n";
  }
  print $fp "x axis info\n";
  print $fp "$xlabel\n";
  print $fp "$nitems\n";
  for ($i=0; $i<$nitems; ++$i) {
    print $fp "$xlabels[$i]\n";
  }
  close($fp);

  $filename = sprintf("out.%s.dat", $filename_string);
  open $fp, "> $filename";
  for ($i=0; $i<$nitems; ++$i) {
    print $fp "$i";
    for ($j=0; $j<$nlines; ++$j) {
      printf $fp "\t%f", $data[$j][$i];
    }
    print $fp "\n";
  }
  close($fp);
}
#
#  Plot functions for a list of 2d x/y graphs
#
#  read_plot2d: companion to the C print_plot2d, reads what C outputted
#  the outputs are sent in an array, the only input is the datinfo file
#  name.
#
sub read_plot2d {
  my $datinfo_prefix = shift @_;
  $datinfo_prefix =~ s/.datinfo$//;

  my $nlines;
  my @useline;
  my $nitems;
  my @data;
  my $lower_is_better;
  my $percentage_away_to_flag;
  my $xlabel;
  my @xlabels;
  my $ylabel;
  my @ylabels;
  my $graph_title;
  my $fstring;

  my $i, $j;
  my $filename;
  my $fp;
  my $tmp;

  $filename = sprintf("%s.datinfo", $datinfo_prefix);
  open $fp, "< $filename";
  $tmp = <$fp>; # 2d
  $fstring = <$fp>; chomp $fstring;
  $graph_title = <$fp>; chomp $graph_title;
  $tmp = <$fp>; # misc info
  $lower_is_better = <$fp>; chomp $lower_is_better;
  $percentage_away_to_flag = <$fp>; chomp $percentage_away_to_flag;
  $tmp = <$fp>; # y axis info
  $ylabel = <$fp>; chomp $ylabel;
  $nlines = <$fp>; chomp $nlines;
  for ($i=0; $i<$nlines; ++$i) {
    $ylabels[$i] = <$fp>;  chomp $ylabels[$i];
    $useline[$i] = <$fp>;  chomp $useline[$i];
  }
  $tmp = <$fp>; # x axis info
  $xlabel = <$fp>; chomp $xlabel;
  $nitems = <$fp>; chomp $nitems;
  for ($i=0; $i<$nitems; ++$i) {
    $xlabels[$i] = <$fp>;  chomp $xlabels[$i];
  }
  close($fp);

  $filename = sprintf("%s.dat", $datinfo_prefix);
  open $fp, "< $filename";
  for ($i=0; $i<$nitems; ++$i) {
    $line = <$fp>; chomp $line; # i data data data ..
    @line = split(/\t/, $line);
    for ($j=0; $j<$nlines; ++$j) {
      $data[$j][$i] = $line[$j + 1];
    }
  }
  close($fp);

  return ($nlines, \@useline, $nitems, \@data,
    $lower_is_better, $percentage_away_to_flag,
    $xlabel, \@xlabels, $ylabel, \@ylabels,
    $graph_title, $fstring);
}
#  Output is
#  - plotme.$filename_string.gnuplot
#  - plotme.$filename_string.dat
#  --> above generate graph.$filename_string.svg via gnuplot
sub mkgnuplot_plot2d {
  my $nlines = shift @_;
  my @useline = @{shift @_};
  my $nitems = shift @_;
  my @data = @{shift @_};
  my $lower_is_better = shift @_;
  my $negate_if_needed = shift @_;
  my $xlabel = shift @_;
  my @xlabels = @{shift @_};
  my $ylabel = shift @_;
  my @ylabels = @{shift @_};
  my $graph_title = shift @_;
  my $filename_string = shift @_;

  my $i, $j;
  my $filename;
  my $fp;
  my $tmp;
  my $smallest_data;
  my $largest_data;
  my $hpixels;
  my $vpixels;
  my $use_boxes = 0;
  my $negate_it = 0;

  if (!$lower_is_better && $negate_if_needed) {
    $negate_it = 1;
  }

# set horizontal pixels for svg to at least 3 x nhosts
  $hpixels = 800;
  if ($nitems * 3 > $hpixels) {
    $hpixels = 3 * $nitems;
  }
  $vpixels = $hpixels / 3;
  if ($vpixels > 850) {
    $vpixels = 850;
  }

  $largest_data = $data[0][0];
  $smallest_data = $data[0][0];
  for ($i=0; $i<$nlines; ++$i) {
    for ($j=0; $j<$nitems; ++$j) {
      if ($data[$i][$j] > $largest_data) {
        $largest_data = $data[$i][$j];
      }
      if ($data[$i][$j] < $smallest_data) {
        $smallest_data = $data[$i][$j];
      }
    }
  }
  if ($negate_it) {
    $tmp = $largest_data;
    $largest_data = -1 * $smallest_data;
    $smallest_data = -1 * $tmp;
    $ylabel = "Negation of $ylabel, so high is bad";
  }
  $extra_data_range_for_y_axis = ($largest_data-$smallest_data)*0.02;
  if ($extra_data_range_for_y_axis < 0.001) {
    $extra_data_range_for_y_axis = 0.001;
  }
  $filename = sprintf("plotme.%s.gnuplot", $filename_string);
  open $fp, "> $filename";
  printf $fp (
    "# To produce a .svg, use the 'set terminal' and 'set output'\n"
  . "# lines below, and set the svg size to whatever you want (and\n"
  . "# leave the pause -1 commented).  For interactive use comment\n"
  . "# out the terminal/output lines and uncomment the pause.\n"
  . "# (Note: some gnuplot programs control svg size via \"size #,#\"\n"
  . "# and others use \"picsize # #\")\n"
  . "# ----------------------------------------------------\n"
  . "set terminal svg \n"
  . "set output \"graph.$filename_string.svg\"\n"
  . "# ----------------------------------------------------\n"
  . "set key outside \n"
  . "set xlabel \"%s\"\n"
  . "set ylabel \"%s\"\n"
  . "set title \"%s\"\n"
  . "set tics out\n"
  . "set yrange [%f:%f]\n"
  . "set boxwidth 0.92\n"
  . "set tics out\n"
  . "plot \\\n",
    $xlabel, $ylabel, $graph_title,
    $smallest_data - $extra_data_range_for_y_axis,
    $largest_data + $extra_data_range_for_y_axis
  );
  for ($i=0; $i<$nlines; ++$i) {
    $withwhat = "linespoints";
    if ($use_boxes) { $withwhat = "boxes"; }
    if ($useline[$i]) {
      printf $fp (
        "  \"plotme.%s.dat\" using 1:%d title \"%s\" with $withwhat",
        $filename_string, $i + 2, $ylabels[$i]
      );
      if ($i < $nlines-1) { printf $fp (", \\"); }
      printf $fp "\n";
    }
  }
  printf $fp "# pause -1\n\n";

  printf $fp ("# FYI, x-axis (%s) labels are:\n", $xlabel);
  for ($i=0; $i<$nitems; ++$i) {
    printf $fp ("# - idx %d: %s\n", $i, $xlabels[$i]);
  }
  close($fp);
	
  $filename = sprintf("plotme.%s.dat", $filename_string);
  open($fp, "> $filename");
  for ($j=0; $j<$nitems; ++$j) {
    printf $fp ("%d", $j);
    for ($i=0; $i<$nlines; ++$i) {
      if ($negate_it) {
        printf $fp (" %f", -1 * $data[$i][$j]);
      } else {
        printf $fp (" %f", $data[$i][$j]);
      }
    }
    printf $fp ("\n");
  }
  close($fp);
}
#
#  Plot functions for a 3d x/y graph
#
#  read_plot3d: companion to the C print_plot3d, reads what C outputted
#  the outputs are sent in an array, the only input is the datinfo file
#  name.
#
sub read_plot3d {
  my $datinfo_prefix = shift @_;
  $datinfo_prefix =~ s/.datinfo$//;

  my $ny;
  my @useline;
  my $nx;
  my @data;
  my $xlabel;
  my @xlabels;
  my $ylabel;
  my @ylabels;
  my $graph_title;
  my $fstring;

  my $i, $j;
  my $filename;
  my $fp;
  my $tmp;

  $filename = sprintf("%s.datinfo", $datinfo_prefix);
  open $fp, "< $filename";
  $tmp = <$fp>; # 3d
  $fstring = <$fp>; chomp $fstring;
  $graph_title = <$fp>; chomp $graph_title;
  $tmp = <$fp>; # y axis info
  $ylabel = <$fp>; chomp $ylabel;
  $ny = <$fp>; chomp $ny;
  for ($i=0; $i<$ny; ++$i) {
    $ylabels[$i] = <$fp>;  chomp $ylabels[$i];
  }
  $tmp = <$fp>; # x axis info
  $xlabel = <$fp>; chomp $xlabel;
  $nx = <$fp>; chomp $nx;
  for ($i=0; $i<$nx; ++$i) {
    $xlabels[$i] = <$fp>;  chomp $xlabels[$i];
  }
  close($fp);

  $filename = sprintf("%s.dat", $datinfo_prefix);
  open $fp, "< $filename";
  for ($i=0; $i<$nx; ++$i) {
    $line = <$fp>; chomp $line; # i data data data ..
    @line = split(/\t/, $line);
    for ($j=0; $j<$ny; ++$j) {
      $data[$i][$j] = $line[$j + 1];
    }
  }
  close($fp);

  return ($nx, $ny, \@data,
    $xlabel, \@xlabels, $ylabel, \@ylabels,
    $graph_title, $fstring);
}
#  Output is
#  - plotme.$filename_string.gnuplot
#  - plotme.$filename_string.dat
#  --> above generate graph.$filename_string.svg via gnuplot
sub mkgnuplot_plot3d {
  my $nx = shift @_;
  my $ny = shift @_;
  my @data = @{shift @_};
  my $xlabel = shift @_;
  my @xlabels = @{shift @_};
  my $ylabel = shift @_;
  my @ylabels = @{shift @_};
  my $graph_title = shift @_;
  my $filename_string = shift @_;

  my $i, $j;
  my $filename;
  my $fp;
  my $tmp;
  my $smallest_data;
  my $largest_data;
  my $hpixels;
  my $vpixels;
  my $use_boxes = 0;

# set pixels for svg to at least 3 x nhosts
  $hpixels = 800;
  if ($nx * 3 > $hpixels) {
    $hpixels = 3 * $nx;
  }
  $vpixels = 800;
  if ($ny * 3 > $vpixels) {
    $vpixels = 3 * $ny;
  }

  $largest_data = $data[0][0];
  $smallest_data = $data[0][0];
  for ($i=0; $i<$nx; ++$i) {
    for ($j=0; $j<$ny; ++$j) {
      if ($data[$i][$j] > $largest_data) {
        $largest_data = $data[$i][$j];
      }
      if ($data[$i][$j] < $smallest_data) {
        $smallest_data = $data[$i][$j];
      }
    }
  }

  $filename = sprintf("plotme.%s.gnuplot", $filename_string);
  open $fp, "> $filename";
  printf $fp (
    "# To produce a .svg, use the 'set terminal' and 'set output'\n"
  . "# lines below, and set the svg size to whatever you want (and\n"
  . "# leave the pause -1 commented).  For interactive use comment\n"
  . "# out the terminal/output lines and uncomment the pause.\n"
  . "# (Note: some gnuplot programs control svg size via \"size #,#\"\n"
  . "# and others use \"picsize # #\")\n"
  . "# ----------------------------------------------------\n"
  . "set terminal svg\n"
  . "set output \"graph.$filename_string.svg\"\n"
  . "# ----------------------------------------------------\n"
  . "set tics out\n"
  . "set view map\n"
  . "set pm3d corners2color c1\n"
  . "set title \"$graph_title\"\n"
  . "splot \"plotme.$filename_string.dat\" using 1:2:3 with pm3d\n"
  . "# pause -1\n");
  printf $fp ("# FYI, x-axis (%s) labels are:\n", $xlabel);
  for ($i=0; $i<$nx; ++$i) {
    printf $fp ("# - idx %d: %s\n", $i, $xlabels[$i]);
  }
  close($fp);
	
  $filename = sprintf("plotme.%s.dat", $filename_string);
  open($fp, "> $filename");
  for ($i=0; $i<$nx; ++$i) {
    for ($j=0; $j<$ny; ++$j) {
      printf $fp ("%d %d %f\n", $i, $j, $data[$i][$j]);
    }
    printf $fp ("%d %d 0\n\n", $i, $j);
  }
  for ($j=0; $j<$ny + 1; ++$j) {
    printf $fp ("%d %d 0\n", $i, $j);
  }
  close($fp);
}
#  Helper to run the above two functions together and gnuplot the output.
#  input:
#    $datinfo_prefix
sub mkreport_plot3d {
  my $datinfo_prefix = shift @_;

  my $nx;
  my @ny;
  my @data;
  my $xlabel;
  my @xlabels;
  my $ylabel;
  my @ylabels;
  my $graph_title;
  my $fstring;

  my $data_ref;
  my $xlabels_ref;
  my $ylabels_ref;
  my $i;

  ($nx, $ny, $data_ref,
    $xlabel, $xlabels_ref, $ylabel, $ylabels_ref,
    $graph_title, $fstring) = read_plot3d($datinfo_prefix);
  system "rm -f graph.$fstring.svg";
  @data = @{$data_ref};
  @xlabels = @{$xlabels_ref};
  @ylabels = @{$ylabels_ref};

  mkgnuplot_plot3d($nx, $ny, \@data,
    $xlabel, \@xlabels, $ylabel, \@ylabels,
    $graph_title, $fstring);
  system "gnuplot plotme.$fstring.gnuplot";

  if ( -f "extrainfo.$fstring.html" ) {
    print REPORTFH `cat extrainfo.$fstring.html`;
  }

  print REPORTFH "<ul>\n";
  print REPORTFH "<li> <a href=\"graph.$fstring.svg\">" .
                 "[Link to full-sized image]</a>\n";
  system "cp graph.$fstring.svg Report/";
  print REPORTFH "<li> <a href=\"plotme.$fstring.dat.txt\">" .
                 "[gnuplot .dat -- text data]</a>";
  system "cp plotme.$fstring.dat Report/plotme.$fstring.dat.txt";
  print REPORTFH "<li> <a href=\"plotme.$fstring.gnuplot.txt\">" .
                 "[gnuplot .gnuplot -- includes $xlabel labels]</a>";
  system "cp plotme.$fstring.gnuplot Report/plotme.$fstring.gnuplot.txt";
  print REPORTFH "</ul>\n";
  print REPORTFH "<a href=\"graph.$fstring.svg\">\n" .
                 "<img width=\"90%\" src=\"graph.$fstring.svg\">" .
                 "</a>\n";
  print REPORTFH "<hr>\n";

  system "rm -f plotme.$fstring.*";  # should be a .gnuplot and a .dat

  print "  - output is at graph.$fstring.svg (all in 1 graph)\n";
}

#  Identify suspected bad x-values (probably hosts).  Takes a subset
#  of the arguments as the graphing functions above
sub identify_bad_xvalues {
  my $nlines = shift @_;
  my $nitems = shift @_;
  my @data = @{shift @_};
  my $lower_is_better = shift @_;
  my $percentage_away_to_flag = shift @_;
  my @xlabels = @{shift @_};
  my @itemstates;

  my $i;
  my $j;
  my $avg;
  my $stddev;

  for ($j=0; $j<$nitems; ++$j) {
    $itemstates[$j] = 0;
  }

  for ($i=0; $i<$nlines; ++$i) {
    $avg = get_avg(@{$data[$i]});
    $stddev = get_stddev(@{$data[$i]});
    for ($j=0; $j<$nitems; ++$j) {
      if ($lower_is_better) {
        if ($data[$i][$j] > $avg + 4*$stddev ||
            $data[$i][$j] > $avg * (1 + $percentage_away_to_flag/100))
        {
          ++$itemstates[$j];
        }
      } else {
        if ($data[$i][$j] < $avg - 4*$stddev ||
            $data[$i][$j] < $avg / (1 + $percentage_away_to_flag/100))
        {
          ++$itemstates[$j];
        }
      }
    }
  }
  return @itemstates;
}

# similar to the graphing function but writing to table
# write to table.$filename_string.html
sub mktable_plot2d {
  my $nlines = shift @_;
  my @useline = @{shift @_};
  my $nitems = shift @_;
  my @data = @{shift @_};
  my $xlabel = shift @_;
  my @xlabels = @{shift @_};
  my $ylabel = shift @_;
  my @ylabels = @{shift @_};
  my $graph_title = shift @_;
  my $filename_string = shift @_;
  my @itemstates = @{shift @_};

  my $i, $j;
  my $filename;
  my $fp;
  my $tmp;

  $filename = sprintf("table.%s.html", $filename_string);
  open($fp, "> $filename");
  print $fp "<html><body>\n";
  print $fp "<h3> $graphname </h3>\n";
  print $fp "<table border=1>\n";
  print $fp "<tr>\n";
  print $fp "<th colspan=\"2\"> x = $xlabel </th>\n";
  print $fp "<th colspan=\"$nlines\"> y = $ylabel </th>\n";
  print $fp "</tr>\n";
  print $fp "<tr>\n";
  print $fp "<th> $xlabel </th> <th> x </th>\n";
  for ($i=0; $i<$nlines; ++$i) {
    if ($useline[$i]) {
      print $fp "<th> $ylabels[$i] </th>\n";
    } else {
      print $fp "<th> ($ylabels[$i]) </th>\n";
    }
  }
  print $fp "</tr>\n";
  for ($j=0; $j<$nitems; ++$j) {
    print $fp "<tr>\n";
    if ($itemstates[$j] > 0) {
      print $fp "<th> *** $xlabels[$j] *** ($itemstates[$j]) </th> <th> $j </th>\n";
    } else {
      print $fp "<th> $xlabels[$j] </th> <th> $j </th>\n";
    }
    for ($i=0; $i<$nlines; ++$i) {
      print $fp "<td> $data[$i][$j] </td>\n";
    }
    print $fp "</tr>\n";
  }
  print $fp "</table>\n";
  print $fp "</body></html>\n";
  close($fp);
}

#  Helper to run the above two functions together and gnuplot the output.
#  input:
#    $datinfo_prefix
#    produce verbose individual graphs or not
sub mkreport_plot2d {
  my $datinfo_prefix = shift @_;
  my $mkindividual_graphs = shift @_;

  my $nlines;
  my @useline;
  my $nitems;
  my @data;
  my $lower_is_better;
  my $percentage_away_to_flag;
  my $xlabel;
  my @xlabels;
  my $ylabel;
  my @ylabels;
  my $graph_title;
  my $fstring;
  my @itemstates;
  my @badlist;

  my $useline_ref;
  my $data_ref;
  my $xlabels_ref;
  my $ylabels_ref;
  my $i;
  my $j;
  my $tmpstr;
  my $count;

  ($nlines, $useline_ref, $nitems, $data_ref,
    $lower_is_better, $percentage_away_to_flag,
    $xlabel, $xlabels_ref, $ylabel, $ylabels_ref,
    $graph_title, $fstring) = read_plot2d($datinfo_prefix);
  system "rm -f graph.$fstring.svg";
  system "rm -f table.$fstring.html";
  system "rm -f graph.y*.$fstring.svg";

  @useline = @{$useline_ref};
  @data = @{$data_ref};
  @xlabels = @{$xlabels_ref};
  @ylabels = @{$ylabels_ref};

# next steps:
# - make graph with all the lines (useline[] anyway) on one plot
# - do minor statistics
# - make table with all the lines
# - make one graph per line possibly reciprocal
# - print suspicious hosts based on statistics
  mkgnuplot_plot2d($nlines, \@useline, $nitems, \@data,
    $lower_is_better, 0,
    $xlabel, \@xlabels, $ylabel, \@ylabels,
    $graph_title, $fstring);
  system "gnuplot plotme.$fstring.gnuplot";
  system "rm -f plotme.$fstring.*";  # should be a .gnuplot and a .dat

  @itemstates = identify_bad_xvalues($nlines, $nitems, \@data,
    $lower_is_better, $percentage_away_to_flag,
    \@xlabels);

  mktable_plot2d($nlines, \@useline, $nitems, \@data,
    $xlabel, \@xlabels, $ylabel, \@ylabels,
    $graph_title, $fstring, \@itemstates);

  for ($i=0; $i<$nlines; ++$i) {
    $tmpstr = sprintf "y%d.%s", $i + 1, $fstring;
    mkgnuplot_plot2d(1, [1],, $nitems, [$data[$i]],
      $lower_is_better, 1,
      $xlabel, \@xlabels, $ylabel, [$ylabels[$i]],
      $graph_title, $tmpstr);
    system "gnuplot plotme.$tmpstr.gnuplot";
    system "rm -f plotme.$tmpstr.*";
  }

# Create string of suspicious hosts and a set of files "Report/bad.xdata.$x"
# to maintain a list of bad hosts across runs
  @badlist = ();
  for ($i=0; $i<$nitems; ++$i) {
    if ($itemstates[$i]) {
      $tmpstr = "bad.xdata." . $xlabels[$i];
      $tmpstr =~ s/[^A-Za-z0-9._-]//;
      open UTILFH, ">> Report/$tmpstr";
      print UTILFH "test: $fstring\n";
      close UTILFH;
      @badlist = (@badlist, $xlabels[$i]);
    }
  }

  print REPORTFH "<ul>\n";
  print REPORTFH "<li> <a href=\"graph.$fstring.svg\">" .
                 "[Link to full-sized image]</a>\n";
  system "cp graph.$fstring.svg Report/";
  print REPORTFH "<li> <a href=\"table.$fstring.html\">" .
                 "[Table of the graphed data, " .
                 "with $xlabel labels included]</a>";
  system "cp table.$fstring.html Report/";
  for ($i=0; $i<$nlines; ++$i) {
    $j = $i + 1;
    $tmpstr = sprintf "y%d.%s", $i + 1, $fstring;
    if ($useline[$i]) {
      print REPORTFH "<li> <a href=\"graph.$tmpstr.svg\">" .
                     "[individualized graph $j]</a> ($ylabels[$i])\n";
    } else {
      print REPORTFH "<li> <a href=\"graph.$tmpstr.svg\">" .
                     "[individualized graph $j]</a> ($ylabels[$i]) " .
                     "(not in main graph)\n";
    }
    system "cp graph.$tmpstr.svg Report/";
    if (!$mkindividual_graphs) {
      system "rm -f graph.$tmpstr.svg"; # user didn't ask for this graph
    }
  }
  if ($#badlist == -1) {
    print REPORTFH "<li> No suspicious hosts found\n";
  } else {
    print REPORTFH "<li> Suspicious hosts list:" . join(", ",@badlist), "\n";
  }
  print REPORTFH "</ul>\n";
  print REPORTFH "<a href=\"graph.$fstring.svg\">\n" .
                 "<img width=\"90%\" src=\"graph.$fstring.svg\">" .
                 "</a>\n";
  print REPORTFH "<hr>\n";

  print "  - output is at\n";
  print "    1. graph.$fstring.svg (all in 1 graph)\n";
  print "    2. table.$fstring.html (see the numbers if you want)\n";
  if ($mkindividual_graphs) {
    print "    3. graph.y*.$fstring.svg (lines graphed individually)\n";
    print "       and inverted if needed so 'bad' datapoints are spikes)\n";
  }
  $count = 0;
  for ($i=0; $i<$nitems; ++$i) {
    if ($itemstates[$i]) { ++$count; }
  }
  if ($count == 0) {
    print "  - No suspicious hosts from dataset $fstring.\n";
  } else {
    print "  - Suspicious hosts from dataset $fstring:\n";
    for ($i=0; $i<$nitems; ++$i) {
      if ($itemstates[$i]) {
        print "    ", $xlabels[$i], "\n";
      }
    }
  }
}

# outermost helper routine for use in mkreport.pl
# Arguments are each optional:
#   out.whatevertest.datinfo  (that .datinfo and a .dat should exist)
#   -individual
# With an additional piece of context it makes use of being the existance
# of a file like header1.html etc.  If no such file exists we suppose
# we're running outside of the ClusterTester harness and produce slightly
# more generic output:
#   - no header describing the testcase
#   - no assumption about what .datinfo to process, only what's on cmdline
# but if the file does exist then we suppose we're in some
# ClusterTester/test*/ directory and
#   - include header#.html in the report
#   - if no .datinfo is specified, all the .datinfo in the current
#     directory are used
sub mkreport_main {
  my @argvin = @{shift @_};
  my $datinfo_prefix;
  my @datinfo_prefixes;
  my $mkindividual_graphs;
  my $line;
  my $line3;
  my $fp;
  my $arg;
  my $tmp;
  my $x;
  my $i;
  my $testnum;

# detect if we're in the ClusterTester harness or not
  $testnum = 0;  # will mean no header, not in ClusterTester harness
  for ($i=1; $i<10; ++$i) {
    if ( -f "header$i.html" ) {
      $testnum = $i;
    }
  }
  if ($testnum > 0 && ! -f "../support/header.html") {
    $testnum = 0;
  }

  system "rm -rf Report";
  system "mkdir -p Report";

  $mkindividual_graphs = 0;
  @datinfo_prefixes = ();
  for $arg (@argvin) {
    if ($arg eq "-individual") { $mkindividual_graphs = 1; }
    else {
      $tmp = $arg;
      $tmp =~ s/\.datinfo$//;
      @datinfo_prefixes = (@datinfo_prefixes, $tmp);
    }
  }
# if @datinfo_prefixes is empty, auto-detect and put a lot of
# extra work into what order to show the files in since
#   some_file_4
#   some_file_200000
# are most likely wanted to be seen in that order.
  if ((! @datinfo_prefixes) && ($testnum != "")) {
    opendir UTILDIR, ".";
    while ($x = readdir UTILDIR) {
      if ($x =~ /(.*)\.datinfo$/) {
        @datinfo_prefixes = (@datinfo_prefixes, $1);
      }
    }
    closedir UTILDIR;
  }
  sub myFnameSort {
    my $tmp;
    my $i;
    my @lista;
    my @listb;
    @lista = ();
    $tmp = $a;
    while ($tmp ne "") {
      if ($tmp =~ s/^([^0-9]+)//) { @lista = (@lista, $1); }
      if ($tmp =~ s/^([0-9]+)//)  { @lista = (@lista, $1); }
    }
    @listb = ();
    $tmp = $b;
    while ($tmp ne "") {
      if ($tmp =~ s/^([^0-9]+)//) { @listb = (@listb, $1); }
      if ($tmp =~ s/^([0-9]+)//)  { @listb = (@listb, $1); }
    }
    $i = 0;
    while ($i <= $#lista && $i <= $#listb) {
      if ($lista[$i] =~ /[0-9]/ && $listb[$i] =~ /[0-9]/) {
        if ($lista[$i] < $listb[$i]) { return -1; }
        if ($lista[$i] > $listb[$i]) { return 1; }
      }
      if ($lista[$i] lt $listb[$i]) { return -1; }
      if ($lista[$i] gt $listb[$i]) { return 1; }
      ++$i;
    }
    if ($i <= $#lista) { return 1; }
    if ($i <= $#listb) { return -1; }
    return 0;
  }
  @datinfo_prefixes = sort myFnameSort @datinfo_prefixes;
  if (! @datinfo_prefixes) {
    print "mkreport:  no .datinfo files found\n";
    if ($testnum > 0) {
      open REPORTFH, "> Report/report${testnum}.html";
    } else {
      open REPORTFH, "> Report/report.html";
    }
    print REPORTFH "No .datinfo files found\n";
    close REPORTFH;
    exit(1);
  }

  if ($testnum > 0) {
    open REPORTFH, "> Report/report${testnum}.html";
    print REPORTFH `cat ../support/header.html`;
    print REPORTFH `cat header$testnum.html`;
  } else {
    open REPORTFH, "> Report/report.html";
    print REPORTFH "<html><body>\n";
  }

  for $datinfo_prefix (@datinfo_prefixes) {
    if ( ! -f "$datinfo_prefix.datinfo" ) {
      print "mkreport:  $datinfo_prefix.datinfo not found\n";
      exit(1);
    }

    open $fp, "< $datinfo_prefix.datinfo";
    $line = <$fp>; chomp $line;
    $line3 = <$fp>; chomp $line3;
    $line3 = <$fp>; chomp $line3;
    close $fp;

    print "Parsing data from $datinfo_prefix.datinfo\n";
    print REPORTFH "<h3> Graph from dataset $datinfo_prefix -- $line3</h3>\n";

    if ($line eq "2d") {
      mkreport_plot2d($datinfo_prefix, $mkindividual_graphs);
    } elsif ($line eq "3d") {
      mkreport_plot3d($datinfo_prefix);
    }
  }
  print REPORTFH "</body></html>\n";
  close REPORTFH;
}
